from .rulesets import RuleSet


permissions = RuleSet()


def add_perm(name, pred):
    permissions.add_rule(name, pred)


def set_perm(name, pred):
    permissions.set_rule(name, pred)


def remove_perm(name):
    permissions.remove_rule(name)


def perm_exists(name):
    return permissions.rule_exists(name)


def has_perm(name, *args, **kwargs):
    return permissions.test_rule(name, *args, **kwargs)


class ObjectPermissionBackend(object):
    def authenticate(self, *args, **kwargs):
        return None

    def has_perm(self, user, perm, *args, **kwargs):
        return has_perm(perm, user, *args, **kwargs)

    def has_module_perms(self, user, app_label):
        return has_perm(app_label, user)



from .predicates import predicate


class RuleSet(dict):
    def test_rule(self, name, *args, **kwargs):
        return name in self and self[name].test(*args, **kwargs)

    def rule_exists(self, name):
        return name in self

    def add_rule(self, name, pred):
        if name in self:
            raise KeyError('A rule with name `%s` already exists' % name)
        self[name] = pred

    def set_rule(self, name, pred):
        self[name] = pred

    def remove_rule(self, name):
        del self[name]

    def __setitem__(self, name, pred):
        fn = predicate(pred)
        super(RuleSet, self).__setitem__(name, fn)


# Shared rule set

default_rules = RuleSet()


def add_rule(name, pred):
    default_rules.add_rule(name, pred)


def set_rule(name, pred):
    default_rules.set_rule(name, pred)


def remove_rule(name):
    default_rules.remove_rule(name)


def rule_exists(name):
    return default_rules.rule_exists(name)


def test_rule(name, *args, **kwargs):
    return default_rules.test_rule(name, *args, **kwargs)



from __future__ import absolute_import

import sys
import unittest

from django.contrib.auth.models import AnonymousUser
from django.core.exceptions import ImproperlyConfigured
from django import http
from django.test import TestCase
from rest_framework.decorators import action
from rest_framework.response import Response
from rest_framework.serializers import ModelSerializer
from rest_framework.test import APIRequestFactory
from rest_framework.viewsets import ModelViewSet

import rules
from rules.contrib.rest_framework import AutoPermissionViewSetMixin


@unittest.skipIf(sys.version_info.major < 3, "Python 3 only")
class AutoPermissionRequiredMixinTests(TestCase):
    def setUp(self):
        from testapp.models import TestModel

        class TestModelSerializer(ModelSerializer):
            class Meta:
                model = TestModel
                fields = "__all__"

        class TestViewSet(AutoPermissionViewSetMixin, ModelViewSet):
            queryset = TestModel.objects.all()
            serializer_class = TestModelSerializer
            permission_type_map = AutoPermissionViewSetMixin.permission_type_map.copy()
            permission_type_map["custom_detail"] = "add"
            permission_type_map["custom_nodetail"] = "add"

            @action(detail=True)
            def custom_detail(self, request):
                return Response()

            @action(detail=False)
            def custom_nodetail(self, request):
                return Response()

            @action(detail=False)
            def unknown(self, request):
                return Response()

        self.model = TestModel
        self.vs = TestViewSet
        self.req = APIRequestFactory().get("/")
        self.req.user = AnonymousUser()

    def test_predefined_action(self):
        # Create should be allowed due to the add permission set on TestModel
        self.assertEqual(self.vs.as_view({"get": "create"})(self.req).status_code, 201)
        # List should be allowed due to None in permission_type_map
        self.assertEqual(
            self.vs.as_view({"get": "list"})(self.req, pk=1).status_code, 200
        )
        # Retrieve should be allowed due to the view permission set on TestModel
        self.assertEqual(
            self.vs.as_view({"get": "retrieve"})(self.req, pk=1).status_code, 200
        )
        # Destroy should be forbidden due to missing delete permission
        self.assertEqual(
            self.vs.as_view({"get": "destroy"})(self.req, pk=1).status_code, 403
        )

    def test_custom_actions(self):
        # Both should not produce 403 due to being mapped to the add permission
        self.assertEqual(
            self.vs.as_view({"get": "custom_detail"})(self.req, pk=1).status_code, 404
        )
        self.assertEqual(
            self.vs.as_view({"get": "custom_nodetail"})(self.req).status_code, 200
        )

    def test_unknown_action(self):
        with self.assertRaises(ImproperlyConfigured):
            self.vs.as_view({"get": "unknown"})(self.req)



from __future__ import absolute_import

import sys

from django.conf import settings
from django.db import models

try:
    from django.utils.encoding import python_2_unicode_compatible
except ImportError:
    def python_2_unicode_compatible(c):
        return c

import rules


@python_2_unicode_compatible
class Book(models.Model):
    isbn = models.CharField(max_length=50, unique=True)
    title = models.CharField(max_length=100)
    author = models.ForeignKey(settings.AUTH_USER_MODEL, on_delete=models.CASCADE)

    def __str__(self):
        return self.title


if sys.version_info.major >= 3:
    from rules.contrib.models import RulesModel

    class TestModel(RulesModel):
        class Meta:
            rules_permissions = {"add": rules.always_true, "view": rules.always_true}

        @classmethod
        def preprocess_rules_permissions(cls, perms):
            perms["custom"] = rules.always_true



from django.core.exceptions import ImproperlyConfigured, PermissionDenied


class AutoPermissionViewSetMixin:
    """
    Enforces object-level permissions in ``rest_framework.viewsets.ViewSet``,
    deriving the permission type from the particular action to be performed..

    As with ``rules.contrib.views.AutoPermissionRequiredMixin``, this only works when
    model permissions are registered using ``rules.contrib.models.RulesModelMixin``.
    """

    # Maps API actions to model permission types. None as value skips permission
    # checks for the particular action.
    # This map needs to be extended when custom actions are implemented
    # using the @action decorator.
    # Extend or replace it in subclasses like so:
    # permission_type_map = {
    #     **AutoPermissionViewSetMixin.permission_type_map,
    #     "close": "change",
    #     "reopen": "change",
    # }
    permission_type_map = {
        "create": "add",
        "destroy": "delete",
        "list": None,
        "partial_update": "change",
        "retrieve": "view",
        "update": "change",
    }

    def initial(self, *args, **kwargs):
        """Ensures user has permission to perform the requested action."""
        super().initial(*args, **kwargs)

        if not self.request.user:
            # No user, don't check permission
            return

        # Get the handler for the HTTP method in use
        try:
            if self.request.method.lower() not in self.http_method_names:
                raise AttributeError
            handler = getattr(self, self.request.method.lower())
        except AttributeError:
            # method not supported, will be denied anyway
            return

        try:
            perm_type = self.permission_type_map[self.action]
        except KeyError:
            raise ImproperlyConfigured(
                "AutoPermissionViewSetMixin tried to authorize a request with the "
                "{!r} action, but permission_type_map only contains: {!r}".format(
                    self.action, self.permission_type_map
                )
            )
        if perm_type is None:
            # Skip permission checking for this action
            return

        # Determine whether we've to check object permissions (for detail actions)
        obj = None
        extra_actions = self.get_extra_actions()
        # We have to access the unbound function via __func__
        if handler.__func__ in extra_actions:
            if handler.detail:
                obj = self.get_object()
        elif self.action not in ("create", "list"):
            obj = self.get_object()

        # Finally, check permission
        perm = self.get_queryset().model.get_perm(perm_type)
        if not self.request.user.has_perm(perm, obj):
            raise PermissionDenied



from .rulesets import RuleSet, add_rule, set_rule, remove_rule, rule_exists, test_rule
from .permissions import add_perm, set_perm, remove_perm, perm_exists, has_perm
from .predicates import (Predicate, predicate, always_true, always_false,
                         always_allow, always_deny, is_authenticated,
                         is_superuser, is_staff, is_active, is_group_member)

VERSION = (2, 2, 0, 'final', 1)

default_app_config = 'rules.apps.RulesConfig'



from os.path import abspath, dirname

BASE_DIR = dirname(dirname(abspath(__file__)))

DEBUG = True

ADMINS = [
    ('test@example.com', 'Administrator'),
]

DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.sqlite3',
        'NAME': ':memory:',
    },
}

INSTALLED_APPS = [
    'django.contrib.admin',
    'django.contrib.auth',
    'django.contrib.contenttypes',
    'django.contrib.messages',
    'django.contrib.sessions',
    'rules',
    'testapp',
]

MIDDLEWARE = [
    'django.contrib.sessions.middleware.SessionMiddleware',
    'django.contrib.auth.middleware.AuthenticationMiddleware',
    'django.contrib.messages.middleware.MessageMiddleware',
]

AUTHENTICATION_BACKENDS = [
    'rules.permissions.ObjectPermissionBackend',
    'django.contrib.auth.backends.ModelBackend',
]

CACHE_BACKEND = 'locmem://'

SECRET_KEY = 'thats-a-secret'

ROOT_URLCONF = 'testapp.urls'

TEMPLATES = [
    {
        'BACKEND': 'django.template.backends.django.DjangoTemplates',
        'DIRS': [],
        'APP_DIRS': True,
        'OPTIONS': {
            'debug': DEBUG,
            'context_processors': [
                'django.template.context_processors.debug',
                'django.template.context_processors.request',
                'django.contrib.auth.context_processors.auth',
                'django.contrib.messages.context_processors.messages',
            ],
        },
    },
]



from functools import wraps

from django.conf import settings
from django.contrib.auth import REDIRECT_FIELD_NAME
from django.contrib.auth import mixins
from django.contrib.auth.views import redirect_to_login
from django.core.exceptions import PermissionDenied, ImproperlyConfigured, FieldError
from django.shortcuts import get_object_or_404
from django.utils.encoding import force_text
from django.views.generic import CreateView, DeleteView, DetailView, UpdateView


# These are made available for convenience, as well as for use in Django
# versions before 1.9. For usage help see Django's docs for 1.9 or later.
from django.views.generic.edit import BaseCreateView

from rules.compat.six import string_types, wraps

LoginRequiredMixin = mixins.LoginRequiredMixin
UserPassesTestMixin = mixins.UserPassesTestMixin


class PermissionRequiredMixin(mixins.PermissionRequiredMixin):
    """
    CBV mixin to provide object-level permission checking to views. Best used
    with views that inherit from ``SingleObjectMixin`` (``DetailView``,
    ``UpdateView``, etc.), though not required.

    The single requirement is for a ``get_object`` method to be available
    in the view. If there's no ``get_object`` method, permission checking
    is model-level, that is exactly like Django's ``PermissionRequiredMixin``.
    """
    def get_permission_object(self):
        """
        Override this method to provide the object to check for permission
        against. By default uses ``self.get_object()`` as provided by
        ``SingleObjectMixin``. Returns None if there's no ``get_object``
        method.
        """
        if not isinstance(self, BaseCreateView):
            # We do NOT want to call get_object in a BaseCreateView, see issue #85
            if hasattr(self, 'get_object') and callable(self.get_object):
                # Requires SingleObjectMixin or equivalent ``get_object`` method
                return self.get_object()
        return None

    def has_permission(self):
        obj = self.get_permission_object()
        perms = self.get_permission_required()
        return self.request.user.has_perms(perms, obj)


class AutoPermissionRequiredMixin(PermissionRequiredMixin):
    """
    An extended variant of PermissionRequiredMixin which automatically determines
    the permission to check based on the type of view it's used with.

    It works by checking the current view for being an instance of a pre-defined
    list of view types. On a match, the corresponding permission type (such as
    "add" or "change") is converted into the full model-specific permission name
    and checked. See the permission_type_map attribute for the default view type ->
    permission type mappings.

    When a view using this mixin has an attribute ``permission_type``, that type
    is used directly and overwrites the permission_type_map for the particular
    view. A permission type of ``None`` (either as ``permission_type`` or in
    ``permission_type_map``) causes permission checking to be skipped. If the type
    of permission to check for should depend on dynamic factors other than the view
    type, you may overwrite the ``permission_type`` attribute with a ``@property``.

    The ``permission_required`` attribute behaves like it does in
    ``PermissionRequiredMixin`` and can be used to specify concrete permission name(s)
    to be checked in addition to the automatically derived one.

    NOTE: The model-based permission registration from ``rules.contrib.models``
    must be used with the models for which you create views using this mixin,
    because the permission names are derived via ``RulesModelMixin.get_perm()``
    internally. The second requirement is the presence of either an attribute
    ``model`` holding the ``Model`` the view acts on, or the ``get_queryset()``
    method as provided by Django's ``SingleObjectMixin``. Hence with the normal
    model views, you don't need to care about anything.
    """

    # These reflect Django's default model permissions. If needed, this list can be
    # extended or replaced entirely when subclassing, like so:
    # permission_type_map = [
    #     (SomeCustomViewType, "add"),
    #     (SomeOtherCustomViewType, "some_fancy_action"),
    #     *AutoPermissionRequiredMixin.permission_type_map,
    # ]
    # Note that ordering matters, which is why this is a list and not a dict. The
    # first entry for which isinstance(self, view_type) returns True will be used.
    permission_type_map = [
        (CreateView, "add"),
        (UpdateView, "change"),
        (DeleteView, "delete"),
        (DetailView, "view"),
    ]

    def get_permission_required(self):
        """Adds the correct permission to check according to view type."""
        try:
            perm_type = self.permission_type
        except AttributeError:
            # Perform auto-detection by view type
            for view_type, _perm_type in self.permission_type_map:
                if isinstance(self, view_type):
                    perm_type = _perm_type
                    break
            else:
                raise ImproperlyConfigured(
                    "AutoPermissionRequiredMixin was used, but permission_type was "
                    "neither set nor could be determined automatically for {0}. "
                    "Consider setting permission_type on the view manually or "
                    "adding {0} to the permission_type_map."
                    .format(self.__class__.__name__)
                )

        perms = []
        if perm_type is not None:
            model = getattr(self, "model", None)
            if model is None:
                model = self.get_queryset().model
            perms.append(model.get_perm(perm_type))

        # If additional permissions have been defined, consider them as well
        if self.permission_required is not None:
            perms.extend(super().get_permission_required())
        return perms


def objectgetter(model, attr_name='pk', field_name='pk'):
    """
    Helper that returns a function suitable for use as the ``fn`` argument
    to the ``permission_required`` decorator. Internally uses
    ``get_object_or_404``, so keep in mind that this may raise ``Http404``.

    ``model`` can be a model class, manager or queryset.

    ``attr_name`` is the name of the view attribute.

    ``field_name`` is the model's field name by which the lookup is made, eg.
    "id", "slug", etc.
    """
    def _getter(request, *view_args, **view_kwargs):
        if attr_name not in view_kwargs:
            raise ImproperlyConfigured(
                'Argument {0} is not available. Given arguments: [{1}]'
                .format(attr_name, ', '.join(view_kwargs.keys())))
        try:
            return get_object_or_404(model, **{field_name: view_kwargs[attr_name]})
        except FieldError:
            raise ImproperlyConfigured(
                'Model {0} has no field named {1}'
                .format(model, field_name))
    return _getter


def permission_required(perm, fn=None, login_url=None, raise_exception=False, redirect_field_name=REDIRECT_FIELD_NAME):
    """
    View decorator that checks for the given permissions before allowing the
    view to execute. Use it like this::

        from django.shortcuts import get_object_or_404
        from rules.contrib.views import permission_required
        from posts.models import Post

        def get_post_by_pk(request, post_id):
            return get_object_or_404(Post, pk=post_id)

        @permission_required('posts.change_post', fn=get_post_by_pk)
        def post_update(request, post_id):
            # ...

    ``perm`` is either a permission name as a string, or a list of permission
    names.

    ``fn`` is an optional callback that receives the same arguments as those
    passed to the decorated view and must return the object to check
    permissions against. If omitted, the decorator behaves just like Django's
    ``permission_required`` decorator, i.e. checks for model-level permissions.

    ``raise_exception`` is a boolean specifying whether to raise a
    ``django.core.exceptions.PermissionDenied`` exception if the check fails.
    You will most likely want to set this argument to ``True`` if you have
    specified a custom 403 response handler in your urlconf. If ``False``,
    the user will be redirected to the URL specified by ``login_url``.

    ``login_url`` is an optional custom URL to redirect the user to if
    permissions check fails. If omitted or empty, ``settings.LOGIN_URL`` is
    used.
    """
    def decorator(view_func):
        @wraps(view_func)
        def _wrapped_view(request, *args, **kwargs):
            # Normalize to a list of permissions
            if isinstance(perm, string_types):
                perms = (perm,)
            else:
                perms = perm

            # Get the object to check permissions against
            if callable(fn):
                obj = fn(request, *args, **kwargs)
            else:  # pragma: no cover
                obj = fn

            # Get the user
            user = request.user

            # Check for permissions and return a response
            if not user.has_perms(perms, obj):
                # User does not have a required permission
                if raise_exception:
                    raise PermissionDenied()
                else:
                    return _redirect_to_login(request, view_func.__name__,
                                              login_url, redirect_field_name)
            else:
                # User has all required permissions -- allow the view to execute
                return view_func(request, *args, **kwargs)

        return _wrapped_view

    return decorator


def _redirect_to_login(request, view_name, login_url, redirect_field_name):
    redirect_url = login_url or settings.LOGIN_URL
    if not redirect_url:  # pragma: no cover
        raise ImproperlyConfigured(
            'permission_required({0}): You must either provide '
            'the "login_url" argument to the "permission_required" '
            'decorator or configure settings.LOGIN_URL'.format(view_name)
        )
    redirect_url = force_text(redirect_url)
    return redirect_to_login(request.get_full_path(), redirect_url, redirect_field_name)



from django.conf.urls import url
from django.contrib import admin

from .views import (change_book, delete_book,
                    view_that_raises, view_with_object, view_with_permission_list,
                    BookUpdateView, BookDeleteView, ViewThatRaises, ViewWithPermissionList,
                    BookUpdateErrorView, BookCreateView)

admin.autodiscover()

urlpatterns = [
    url(r'^admin/', admin.site.urls),

    # Function-based views
    url(r'^(?P<book_id>\d+)/change/$', change_book, name='change_book'),
    url(r'^(?P<book_id>\d+)/delete/$', delete_book, name='delete_book'),
    url(r'^(?P<book_id>\d+)/raise/$', view_that_raises, name='view_that_raises'),
    url(r'^(?P<book_id>\d+)/object/$', view_with_object, name='view_with_object'),
    url(r'^(?P<book_id>\d+)/list/$', view_with_permission_list, name='view_with_permission_list'),

    # Class-based views
    url(r'^cbv/create/$', BookCreateView.as_view(), name='cbv.create_book'),
    url(r'^cbv/(?P<book_id>\d+)/change/$', BookUpdateView.as_view(), name='cbv.change_book'),
    url(r'^cbv/(?P<book_id>\d+)/delete/$', BookDeleteView.as_view(), name='cbv.delete_book'),
    url(r'^cbv/(?P<book_id>\d+)/raise/$', ViewThatRaises.as_view(), name='cbv.view_that_raises'),
    url(r'^cbv/(?P<book_id>\d+)/list/$', ViewWithPermissionList.as_view(), name='cbv.view_with_permission_list'),

    url(r'^cbv/(?P<book_id>\d+)/change-error/$', BookUpdateErrorView.as_view(), name='cbv.change_book_error'),
]



from django import template

from ..rulesets import default_rules


register = template.Library()


@register.simple_tag
def test_rule(name, obj=None, target=None):
    return default_rules.test_rule(name, obj, target)


@register.simple_tag
def has_perm(perm, user, obj=None):
    if not hasattr(user, 'has_perm'):  # pragma: no cover
        return False  # swapped user model that doesn't support permissions
    else:
        return user.has_perm(perm, obj)



from __future__ import absolute_import

from django.http import HttpResponse
from django.views.generic.edit import UpdateView, DeleteView, CreateView

from rules.contrib.views import permission_required, objectgetter
from rules.contrib.views import LoginRequiredMixin, PermissionRequiredMixin

from .models import Book


class BookMixin(object):
    def get_object(self):
        return Book.objects.get(pk=self.kwargs['book_id'])


class BookMixinWithError(object):
    def get_object(self):
        raise AttributeError('get_object')


@permission_required('testapp.change_book', fn=objectgetter(Book, 'book_id'))
def change_book(request, book_id):
    return HttpResponse('OK')



class BookCreateView(LoginRequiredMixin, PermissionRequiredMixin, BookMixin, CreateView):
    fields = ['title']
    template_name = 'empty.html'
    permission_required = 'testapp.create_book'


class BookUpdateView(LoginRequiredMixin, PermissionRequiredMixin, BookMixin, UpdateView):
    fields = ['title']
    template_name = 'empty.html'
    permission_required = 'testapp.change_book'


class BookUpdateErrorView(LoginRequiredMixin, PermissionRequiredMixin, BookMixinWithError, UpdateView):
    fields = ['title']
    template_name = 'empty.html'
    permission_required = 'testapp.change_book'


@permission_required('testapp.delete_book', fn=objectgetter(Book, 'book_id'))
def delete_book(request, book_id):
    return HttpResponse('OK')


class BookDeleteView(LoginRequiredMixin, PermissionRequiredMixin, BookMixin, DeleteView):
    template_name = 'empty.html'
    permission_required = 'testapp.delete_book'


@permission_required('testapp.delete_book', fn=objectgetter(Book, 'book_id'), raise_exception=True)
def view_that_raises(request, book_id):
    return HttpResponse('OK')


class ViewThatRaises(LoginRequiredMixin, PermissionRequiredMixin, BookMixin, DeleteView):
    template_name = 'empty.html'
    raise_exception = True
    permission_required = 'testapp.delete_book'


@permission_required(['testapp.change_book', 'testapp.delete_book'], fn=objectgetter(Book, 'book_id'))
def view_with_permission_list(request, book_id):
    return HttpResponse('OK')


class ViewWithPermissionList(LoginRequiredMixin, PermissionRequiredMixin, BookMixin, DeleteView):
    template_name = 'empty.html'
    permission_required = ['testapp.change_book', 'testapp.delete_book']


@permission_required('testapp.delete_book', fn=objectgetter(Book, 'book_id'))
def view_with_object(request, book_id):
    return HttpResponse('OK')



import sys
import functools
from unittest import TestCase

from rules.predicates import (
    NO_VALUE,
    Predicate,
    always_allow,
    always_deny,
    always_false,
    always_true,
    predicate,
)

if sys.version_info >= (3,):
    from ._test_predicates_kwonly import *


class PredicateTests(TestCase):
    def test_always_true(self):
        assert always_true()

    def test_always_false(self):
        assert not always_false()

    def test_always_allow(self):
        assert always_allow()

    def test_always_deny(self):
        assert not always_deny()

    def test_lambda_predicate(self):
        p = Predicate(lambda x: x == 'a')
        assert p.name == '<lambda>'
        assert p.num_args == 1
        assert p('a')

    def test_lambda_predicate_custom_name(self):
        p = Predicate(lambda x: x == 'a', name='mypred')
        assert p.name == 'mypred'
        assert p.num_args == 1
        assert p('a')

    def test_function_predicate(self):
        def mypred(x):
            return x == 'a'
        p = Predicate(mypred)
        assert p.name == 'mypred'
        assert p.num_args == 1
        assert p('a')

    def test_function_predicate_custom_name(self):
        def mypred(x):
            return x == 'a'
        p = Predicate(mypred, name='foo')
        assert p.name == 'foo'
        assert p.num_args == 1
        assert p('a')

    def test_partial_function_predicate(self):
        def mypred(one, two, three):
            return one < two < three
        p = Predicate(functools.partial(mypred, 1))
        assert p.name == 'mypred'
        assert p.num_args == 2  # 3 - 1 partial
        assert p(2, 3)
        p = Predicate(functools.partial(mypred, 1, 2))
        assert p.name == 'mypred'
        assert p.num_args == 1  # 3 - 2 partial
        assert p(3)

    def test_method_predicate(self):
        class SomeClass(object):
            def some_method(self, arg1, arg2):
                return arg1 == arg2
        obj = SomeClass()
        p = Predicate(obj.some_method)
        assert p.name == 'some_method'
        assert p.num_args == 2
        assert p(2, 2)

    def test_partial_method_predicate(self):
        class SomeClass(object):
            def some_method(self, arg1, arg2):
                return arg1 == arg2
        obj = SomeClass()
        p = Predicate(functools.partial(obj.some_method, 2))
        assert p.name == 'some_method'
        assert p.num_args == 1
        assert p(2)

    def test_class_predicate(self):
        class callableclass(object):
            def __call__(self, arg1, arg2):
                return arg1 == arg2
        fn = callableclass()
        p = Predicate(fn)
        assert p.name == 'callableclass'
        assert p.num_args == 2
        assert p('a', 'a')

    def test_class_predicate_custom_name(self):
        class callableclass(object):
            def __call__(self, arg):
                return arg == 'a'
        fn = callableclass()
        p = Predicate(fn, name='bar')
        assert p.name == 'bar'
        assert p.num_args == 1
        assert p('a')

    def test_predicate_predicate(self):
        def mypred(x):
            return x == 'a'
        p = Predicate(Predicate(mypred))
        assert p.name == 'mypred'
        assert p.num_args == 1
        assert p('a')

    def test_predicate_predicate_custom_name(self):
        def mypred(x):
            return x == 'a'
        p = Predicate(Predicate(mypred, name='foo'))
        assert p.name == 'foo'
        assert p.num_args == 1
        assert p('a')

    def test_predicate_bind(self):
        @predicate(bind=True)
        def is_bound(self):
            return self is is_bound

        assert is_bound()

        p = None

        def mypred(self):
            return self is p

        p = Predicate(mypred, bind=True)
        assert p()

    def test_decorator(self):
        @predicate
        def mypred(arg1, arg2):
            return True
        assert mypred.name == 'mypred'
        assert mypred.num_args == 2

    def test_decorator_noargs(self):
        @predicate()
        def mypred(arg1, arg2):
            return True
        assert mypred.name == 'mypred'
        assert mypred.num_args == 2

    def test_decorator_custom_name(self):
        @predicate('foo')
        def mypred():
            return True
        assert mypred.name == 'foo'
        assert mypred.num_args == 0

        @predicate(name='bar')
        def myotherpred():
            return False
        assert myotherpred.name == 'bar'
        assert myotherpred.num_args == 0

    def test_repr(self):
        @predicate
        def mypred(arg1, arg2):
            return True
        assert repr(mypred).startswith('<Predicate:mypred object at 0x')

    def test_AND(self):
        p_AND1 = always_true & always_false
        assert not p_AND1()
        assert p_AND1.name == '(always_true & always_false)'
        p_AND2 = always_false & always_true
        assert not p_AND2()
        assert p_AND2.name == '(always_false & always_true)'
        p_AND3 = always_true & always_true
        assert p_AND3()
        assert p_AND3.name == '(always_true & always_true)'
        p_AND4 = always_false & always_false
        assert not p_AND4()
        assert p_AND4.name == '(always_false & always_false)'

    def test_OR(self):
        p_OR1 = always_true | always_false
        assert p_OR1()
        assert p_OR1.name == '(always_true | always_false)'
        p_OR2 = always_false | always_true
        assert p_OR2()
        assert p_OR2.name == '(always_false | always_true)'
        p_OR3 = always_true | always_true
        assert p_OR3()
        assert p_OR3.name == '(always_true | always_true)'
        p_OR4 = always_false | always_false
        assert not p_OR4()
        assert p_OR4.name == '(always_false | always_false)'

    def test_XOR(self):
        p_XOR1 = always_true ^ always_false
        assert p_XOR1()
        assert p_XOR1.name == '(always_true ^ always_false)'
        p_XOR2 = always_false ^ always_true
        assert p_XOR2()
        assert p_XOR2.name == '(always_false ^ always_true)'
        p_XOR3 = always_true ^ always_true
        assert not p_XOR3()
        assert p_XOR3.name == '(always_true ^ always_true)'
        p_XOR4 = always_false ^ always_false
        assert not p_XOR4()
        assert p_XOR4.name == '(always_false ^ always_false)'

    def test_INV(self):
        p_INV1 = ~always_true
        assert not p_INV1()
        assert p_INV1.name == '~always_true'
        p_INV2 = ~always_false
        assert p_INV2()
        assert p_INV2.name == '~always_false'
        p_INV3 = ~(~always_true)
        assert p_INV3()
        assert p_INV3.name == 'always_true'
        p_INV4 = ~(~always_false)
        assert not p_INV4()
        assert p_INV4.name == 'always_false'

    def test_var_args(self):
        @predicate
        def p(*args, **kwargs):
            assert len(args) > 0
            assert len(kwargs) == 0
        assert p.num_args == 0
        p.test('a')
        p.test('a', 'b')

    def test_no_args(self):
        @predicate
        def p(*args, **kwargs):
            assert len(args) == 0
            assert len(kwargs) == 0
        assert p.num_args == 0
        p.test()

    def test_one_arg(self):
        @predicate
        def p(a=None, *args, **kwargs):
            assert len(args) == 0
            assert len(kwargs) == 0
            assert a == 'a'
        assert p.num_args == 1
        p.test('a')

    def test_two_args(self):
        @predicate
        def p(a=None, b=None, *args, **kwargs):
            assert len(args) == 0
            assert len(kwargs) == 0
            assert a == 'a'
            assert b == 'b'
        assert p.num_args == 2
        p.test('a', 'b')

    def test_no_mask(self):
        @predicate
        def p(a=None, b=None, *args, **kwargs):
            assert len(args) == 0
            assert len(kwargs) == 1
            'c' in kwargs
            assert a == 'a'
            assert b == 'b'
        p('a', b='b', c='c')

    def test_no_value_marker(self):
        @predicate
        def p(a, b=None):
            assert a == 'a'
            assert b is None

        assert not NO_VALUE
        p.test('a')
        p.test('a', NO_VALUE)

    def test_short_circuit(self):
        @predicate
        def skipped_predicate(self):
            return None

        @predicate
        def shorted_predicate(self):
            raise Exception('this predicate should not be evaluated')

        assert (always_false & shorted_predicate).test() is False
        assert (always_true | shorted_predicate).test() is True

        def raises(pred):
            try:
                pred.test()
                return False
            except Exception as e:
                return 'evaluated' in str(e)

        assert raises(always_true & shorted_predicate)
        assert raises(always_false | shorted_predicate)
        assert raises(skipped_predicate & shorted_predicate)
        assert raises(skipped_predicate | shorted_predicate)

    def test_skip_predicate(self):
        @predicate(bind=True)
        def requires_two_args(self, a, b):
            return a == b if len(self.context.args) > 1 else None

        @predicate
        def passthrough(a):
            return a

        assert (requires_two_args & passthrough).test(True, True) is True
        assert (requires_two_args & passthrough).test(True, False) is False

        # because requires_two_args is called with only one argument
        # its result is not taken into account, only the result of the
        # other predicate matters.
        assert (requires_two_args & passthrough).test(True) is True
        assert (requires_two_args & passthrough).test(False) is False
        assert (requires_two_args | passthrough).test(True) is True
        assert (requires_two_args | passthrough).test(False) is False

        # test that order does not matter
        assert (passthrough & requires_two_args).test(True) is True
        assert (passthrough & requires_two_args).test(False) is False
        assert (passthrough | requires_two_args).test(True) is True
        assert (passthrough | requires_two_args).test(False) is False

        # test that inversion does not modify the result
        assert (~requires_two_args & passthrough).test(True) is True
        assert (~requires_two_args & passthrough).test(False) is False
        assert (~requires_two_args | passthrough).test(True) is True
        assert (~requires_two_args | passthrough).test(False) is False
        assert (passthrough & ~requires_two_args).test(True) is True
        assert (passthrough & ~requires_two_args).test(False) is False
        assert (passthrough | ~requires_two_args).test(True) is True
        assert (passthrough | ~requires_two_args).test(False) is False

        # test that when all predicates are skipped, result is False
        assert requires_two_args.test(True) is False
        assert (requires_two_args | requires_two_args).test(True) is False
        assert (requires_two_args & requires_two_args).test(True) is False

        # test that a skipped predicate doesn't alter the result at all
        assert (requires_two_args | requires_two_args | passthrough).test(True) is True
        assert (requires_two_args & requires_two_args & passthrough).test(True) is True

    def test_invocation_context(self):
        @predicate
        def p1():
            assert id(p1.context) == id(p2.context)
            assert p1.context.args == ('a',)
            return True

        @predicate
        def p2():
            assert id(p1.context) == id(p2.context)
            assert p2.context.args == ('a',)
            return True

        p = p1 & p2
        assert p.test('a')
        assert p.context is None

    def test_invocation_context_nested(self):
        @predicate
        def p1():
            assert p1.context.args == ('b1',)
            return True

        @predicate
        def p2():
            assert p2.context.args == ('b2',)
            return True

        @predicate
        def p():
            assert p1.context.args == ('a',)
            return p1.test('b1') & p2.test('b2')

        assert p.test('a')
        assert p.context is None

    def test_invocation_context_storage(self):
        @predicate
        def p1(a):
            p1.context['p1.a'] = a
            return True

        @predicate
        def p2(a):
            return p2.context['p1.a'] == a

        p = p1 & p2
        assert p.test('a')



from unittest import TestCase

from rules.predicates import predicate


class PredicateKwonlyTests(TestCase):
    def test_predicate_kwargonly(self):
        def p(foo, *, bar):
            return True
        with self.assertRaises(TypeError):
            predicate(p)

        def p2(foo, *a, bar):
            return True
        with self.assertRaises(TypeError):
            predicate(p2)

        def p3(foo, *, bar='bar'):
            return True
        # Should not fail
        predicate(p3)



from django.contrib import admin
from django.contrib.auth import get_permission_codename

from ..permissions import perm_exists

class ObjectPermissionsModelAdminMixin(object):
    def has_view_permission(self, request, obj=None):
        opts = self.opts
        codename = get_permission_codename('view', opts)
        perm = '%s.%s' % (opts.app_label, codename)
        if perm_exists(perm):
            return request.user.has_perm(perm, obj)
        else:
            return self.has_change_permission(request, obj)

    def has_change_permission(self, request, obj=None):
        opts = self.opts
        codename = get_permission_codename('change', opts)
        return request.user.has_perm('%s.%s' % (opts.app_label, codename), obj)

    def has_delete_permission(self, request, obj=None):
        opts = self.opts
        codename = get_permission_codename('delete', opts)
        return request.user.has_perm('%s.%s' % (opts.app_label, codename), obj)


class ObjectPermissionsInlineModelAdminMixin(ObjectPermissionsModelAdminMixin):
    def has_change_permission(self, request, obj=None):  # pragma: no cover
        opts = self.opts
        if opts.auto_created:
            for field in opts.fields:
                if field.rel and field.rel.to != self.parent_model:
                    opts = field.rel.to._meta
                    break
        codename = get_permission_codename('change', opts)
        return request.user.has_perm('%s.%s' % (opts.app_label, codename), obj)

    def has_delete_permission(self, request, obj=None):  # pragma: no cover
        if self.opts.auto_created:
            return self.has_change_permission(request, obj)
        return super(ObjectPermissionsInlineModelAdminMixin, self).has_delete_permission(request, obj)


class ObjectPermissionsModelAdmin(ObjectPermissionsModelAdminMixin, admin.ModelAdmin):
    pass


class ObjectPermissionsStackedInline(ObjectPermissionsInlineModelAdminMixin, admin.StackedInline):
    pass


class ObjectPermissionsTabularInline(ObjectPermissionsInlineModelAdminMixin, admin.TabularInline):
    pass



import logging
import operator
import threading
from functools import partial, update_wrapper

from .compat import inspect


logger = logging.getLogger('rules')


def assert_has_kwonlydefaults(fn, msg):
    argspec = inspect.getfullargspec(fn)
    if hasattr(argspec, 'kwonlyargs'):
        if not argspec.kwonlyargs:
            return
        if not argspec.kwonlydefaults or len(argspec.kwonlyargs) > len(argspec.kwonlydefaults.keys()):
            raise TypeError(msg)


class Context(dict):
    def __init__(self, args):
        super(Context, self).__init__()
        self.args = args


class localcontext(threading.local):
    def __init__(self):
        self.stack = []


_context = localcontext()


class NoValueSentinel(object):
    def __bool__(self):
        return False

    __nonzero__ = __bool__  # python 2


NO_VALUE = NoValueSentinel()

del NoValueSentinel


class Predicate(object):
    def __init__(self, fn, name=None, bind=False):
        # fn can be a callable with any of the following signatures:
        #   - fn(obj=None, target=None)
        #   - fn(obj=None)
        #   - fn()
        assert callable(fn), 'The given predicate is not callable.'
        innerfn = fn
        if isinstance(fn, Predicate):
            fn, num_args, var_args, name = fn.fn, fn.num_args, fn.var_args, name or fn.name
            innerfn = fn
        elif isinstance(fn, partial):
            innerfn = fn.func
            argspec = inspect.getfullargspec(innerfn)
            var_args = argspec.varargs is not None
            num_args = len(argspec.args) - len(fn.args)
            if inspect.ismethod(innerfn):
                num_args -= 1  # skip `self`
            name = fn.func.__name__
        elif inspect.ismethod(fn):
            argspec = inspect.getfullargspec(fn)
            var_args = argspec.varargs is not None
            num_args = len(argspec.args) - 1  # skip `self`
        elif inspect.isfunction(fn):
            argspec = inspect.getfullargspec(fn)
            var_args = argspec.varargs is not None
            num_args = len(argspec.args)
        elif isinstance(fn, object):
            innerfn = getattr(fn, '__call__')
            argspec = inspect.getfullargspec(innerfn)
            var_args = argspec.varargs is not None
            num_args = len(argspec.args) - 1  # skip `self`
            name = name or type(fn).__name__
        else:  # pragma: no cover
            # We handle all cases, so there's no way we can reach here
            raise TypeError('Incompatible predicate.')
        if bind:
            num_args -= 1
        assert_has_kwonlydefaults(innerfn, 'The given predicate is missing defaults for keyword-only arguments')
        assert num_args <= 2, 'Incompatible predicate.'
        self.fn = fn
        self.num_args = num_args
        self.var_args = var_args
        self.name = name or fn.__name__
        self.bind = bind

    def __repr__(self):
        return '<%s:%s object at %s>' % (
            type(self).__name__, str(self), hex(id(self)))

    def __str__(self):
        return self.name

    def __call__(self, *args, **kwargs):
        # this method is defined as variadic in order to not mask the
        # underlying callable's signature that was most likely decorated
        # as a predicate. internally we consistently call ``_apply`` that
        # provides a single interface to the callable.
        if self.bind:
            return self.fn(self, *args, **kwargs)
        return self.fn(*args, **kwargs)

    @property
    def context(self):
        """
        The currently active invocation context. A new context is created as a
        result of invoking ``test()`` and is only valid for the duration of
        the invocation.

        Can be used by predicates to store arbitrary data, eg. for caching
        computed values, setting flags, etc., that can be used by predicates
        later on in the chain.

        Inside a predicate function it can be used like so::

            >>> @predicate
            ... def mypred(a, b):
            ...     value = compute_expensive_value(a)
            ...     mypred.context['value'] = value
            ...     return True
            ...

        Other predicates can later use stored values::

            >>> @predicate
            ... def myotherpred(a, b):
            ...     value = myotherpred.context.get('value')
            ...     if value is not None:
            ...         return do_something_with_value(value)
            ...     else:
            ...         return do_something_without_value()
            ...

        """
        try:
            return _context.stack[-1]
        except IndexError:
            return None

    def test(self, obj=NO_VALUE, target=NO_VALUE):
        """
        The canonical method to invoke predicates.
        """
        args = tuple(arg for arg in (obj, target) if arg is not NO_VALUE)
        _context.stack.append(Context(args))
        logger.debug('Testing %s', self)
        try:
            return bool(self._apply(*args))
        finally:
            _context.stack.pop()

    def __and__(self, other):
        def AND(*args):
            return self._combine(other, operator.and_, args)
        return type(self)(AND, '(%s & %s)' % (self.name, other.name))

    def __or__(self, other):
        def OR(*args):
            return self._combine(other, operator.or_, args)
        return type(self)(OR, '(%s | %s)' % (self.name, other.name))

    def __xor__(self, other):
        def XOR(*args):
            return self._combine(other, operator.xor, args)
        return type(self)(XOR, '(%s ^ %s)' % (self.name, other.name))

    def __invert__(self):
        def INVERT(*args):
            result = self._apply(*args)
            return None if result is None else not result
        if self.name.startswith('~'):
            name = self.name[1:]
        else:
            name = '~' + self.name
        return type(self)(INVERT, name)

    def _combine(self, other, op, args):
        self_result = self._apply(*args)
        if self_result is None:
            return other._apply(*args)

        # short-circuit evaluation
        if op is operator.and_ and not self_result:
            return False
        elif op is operator.or_ and self_result:
            return True

        other_result = other._apply(*args)
        if other_result is None:
            return self_result

        return op(self_result, other_result)

    def _apply(self, *args):
        # Internal method that is used to invoke the predicate with the
        # proper number of positional arguments, inside the current
        # invocation context.
        if self.var_args:
            callargs = args
        elif self.num_args > len(args):
            callargs = args + (None,) * (self.num_args - len(args))
        else:
            callargs = args[:self.num_args]
        if self.bind:
            callargs = (self,) + callargs

        result = self.fn(*callargs)
        result = None if result is None else bool(result)

        logger.debug('  %s = %s', self, 'skipped' if result is None else result)
        return result


def predicate(fn=None, name=None, **options):
    """
    Decorator that constructs a ``Predicate`` instance from any function::

        >>> @predicate
        ... def is_book_author(user, book):
        ...     return user == book.author
        ...

        >>> @predicate(bind=True)
        ... def is_book_author(self, user, book):
        ...     if self.context.args:
        ...         return user == book.author
    """
    if not name and not callable(fn):
        name = fn
        fn = None

    def inner(fn):
        if isinstance(fn, Predicate):
            return fn
        p = Predicate(fn, name, **options)
        update_wrapper(p, fn)
        return p

    if fn:
        return inner(fn)
    else:
        return inner


# Predefined predicates

always_true = predicate(lambda: True, name='always_true')
always_false = predicate(lambda: False, name='always_false')

always_allow = predicate(lambda: True, name='always_allow')
always_deny = predicate(lambda: False, name='always_deny')


def is_bool_like(obj):
    return hasattr(obj, '__bool__') or hasattr(obj, '__nonzero__')


@predicate
def is_authenticated(user):
    if not hasattr(user, 'is_authenticated'):
        return False  # not a user model
    if not is_bool_like(user.is_authenticated):  # pragma: no cover
        # Django < 1.10
        return user.is_authenticated()
    return user.is_authenticated


@predicate
def is_superuser(user):
    if not hasattr(user, 'is_superuser'):
        return False  # swapped user model, doesn't support is_superuser
    return user.is_superuser


@predicate
def is_staff(user):
    if not hasattr(user, 'is_staff'):
        return False  # swapped user model, doesn't support is_staff
    return user.is_staff


@predicate
def is_active(user):
    if not hasattr(user, 'is_active'):
        return False  # swapped user model, doesn't support is_active
    return user.is_active


def is_group_member(*groups):
    assert len(groups) > 0, 'You must provide at least one group name'

    if len(groups) > 3:
        g = groups[:3] + ('...',)
    else:
        g = groups

    name = 'is_group_member:%s' % ','.join(g)

    @predicate(name)
    def fn(user):
        if not hasattr(user, 'groups'):
            return False  # swapped user model, doesn't support groups
        if not hasattr(user, '_group_names_cache'):  # pragma: no cover
            user._group_names_cache = set(user.groups.values_list('name', flat=True))
        return set(groups).issubset(user._group_names_cache)

    return fn



from __future__ import absolute_import

try:
    from inspect import getfullargspec
except ImportError:
    # Python 2 compatibility
    from inspect import getargspec as getfullargspec

from inspect import ismethod, isfunction



from django.apps import AppConfig


class RulesConfig(AppConfig):
    name = 'rules'


class AutodiscoverRulesConfig(RulesConfig):
    def ready(self):
        from django.utils.module_loading import autodiscover_modules
        autodiscover_modules('rules')



from __future__ import absolute_import

import sys
import unittest

from django.core.exceptions import ImproperlyConfigured
from django.test import TestCase

import rules


@unittest.skipIf(sys.version_info.major < 3, "Python 3 only")
class RulesModelTests(TestCase):
    def test_preprocess(self):
        from testapp.models import TestModel

        self.assertTrue(rules.perm_exists("testapp.add_testmodel"))
        self.assertTrue(rules.perm_exists("testapp.custom_testmodel"))

    def test_invalid_config(self):
        from rules.contrib.models import RulesModel

        with self.assertRaises(ImproperlyConfigured):

            class InvalidTestModel(RulesModel):
                class Meta:
                    app_label = "testapp"
                    rules_permissions = "invalid"



import functools
import sys


if sys.version_info[0] == 3:
    string_types = str,
else:
    string_types = basestring,

if sys.version_info[0:2] < (3, 4):
    def wraps(wrapped, assigned=functools.WRAPPER_ASSIGNMENTS,
              updated=functools.WRAPPER_UPDATES):
        def wrapper(f):
            f = functools.wraps(wrapped, assigned, updated)(f)
            f.__wrapped__ = wrapped
            return f
        return wrapper
else:
    wraps = functools.wraps



from __future__ import absolute_import

import sys
import unittest

from django.contrib.auth.models import AnonymousUser
from django.core.exceptions import ImproperlyConfigured, PermissionDenied
from django.http import HttpRequest, Http404
from django.test import RequestFactory, TestCase
from django.urls import reverse
from django.utils.encoding import force_str
from django.views.generic import CreateView, View

import rules
from rules.contrib.views import AutoPermissionRequiredMixin, objectgetter

from testapp.models import Book

from . import TestData


class FBVDecoratorTests(TestData, TestCase):
    def test_objectgetter(self):
        request = HttpRequest()
        book = Book.objects.get(pk=1)

        self.assertEqual(book, objectgetter(Book)(request, pk=1))
        self.assertEqual(book, objectgetter(Book, attr_name="id")(request, id=1))
        self.assertEqual(book, objectgetter(Book, field_name="id")(request, pk=1))

        with self.assertRaises(ImproperlyConfigured):
            # Raise if no `pk` argument is provided to the view
            self.assertEqual(book, objectgetter(Book)(request, foo=1))

        with self.assertRaises(ImproperlyConfigured):
            # Raise if given invalid model lookup field
            self.assertEqual(book, objectgetter(Book, field_name="foo")(request, pk=1))

        with self.assertRaises(Http404):
            # Raise 404 if no model instance found
            self.assertEqual(book, objectgetter(Book)(request, pk=100000))

    def test_permission_required(self):
        # Adrian can change his book
        self.assertTrue(self.client.login(username="adrian", password="secr3t"))
        response = self.client.get(reverse("change_book", args=(1,)))
        self.assertEqual(response.status_code, 200)
        self.assertEqual(force_str(response.content), "OK")

        # Martin can change Adrian's book
        self.assertTrue(self.client.login(username="martin", password="secr3t"))
        response = self.client.get(reverse("change_book", args=(1,)))
        self.assertEqual(response.status_code, 200)
        self.assertEqual(force_str(response.content), "OK")

        # Adrian can delete his book
        self.assertTrue(self.client.login(username="adrian", password="secr3t"))
        response = self.client.get(reverse("delete_book", args=(1,)))
        self.assertEqual(response.status_code, 200)
        self.assertEqual(force_str(response.content), "OK")

        # Martin can *not* create a book
        # Up to Django v2.1, the response was a redirect to login
        self.assertTrue(self.client.login(username="martin", password="secr3t"))
        response = self.client.get(reverse("cbv.create_book"))
        self.assertIn(response.status_code, [302, 403])

        # Martin can *not* delete Adrian's book and is redirected to login
        self.assertTrue(self.client.login(username="martin", password="secr3t"))
        response = self.client.get(reverse("delete_book", args=(1,)))
        self.assertEqual(response.status_code, 302)

        # Martin can *not* delete Adrian's book and an PermissionDenied is raised
        self.assertTrue(self.client.login(username="martin", password="secr3t"))
        response = self.client.get(reverse("view_that_raises", args=(1,)))
        self.assertEqual(response.status_code, 403)

        # Test views that require a list of permissions

        # Adrian has both permissions
        self.assertTrue(self.client.login(username="adrian", password="secr3t"))
        response = self.client.get(reverse("view_with_permission_list", args=(1,)))
        self.assertEqual(response.status_code, 200)
        self.assertEqual(force_str(response.content), "OK")

        # Martin does not have delete permission
        self.assertTrue(self.client.login(username="martin", password="secr3t"))
        response = self.client.get(reverse("view_with_permission_list", args=(1,)))
        self.assertEqual(response.status_code, 302)

        # Test views that accept a static object as argument
        # fn is passed to has_perm as-is

        self.assertTrue(self.client.login(username="adrian", password="secr3t"))
        response = self.client.get(reverse("view_with_object", args=(1,)))
        self.assertEqual(response.status_code, 200)
        self.assertEqual(force_str(response.content), "OK")

        self.assertTrue(self.client.login(username="martin", password="secr3t"))
        response = self.client.get(reverse("view_with_object", args=(1,)))
        self.assertEqual(response.status_code, 302)


class CBVMixinTests(TestData, TestCase):
    def test_get_object_error(self):
        self.assertTrue(self.client.login(username="adrian", password="secr3t"))
        with self.assertRaises(AttributeError):
            self.client.get(reverse("cbv.change_book_error", args=(1,)))

    def test_permission_required_mixin(self):
        # Adrian can change his book
        self.assertTrue(self.client.login(username="adrian", password="secr3t"))
        response = self.client.get(reverse("cbv.change_book", args=(1,)))
        self.assertEqual(response.status_code, 200)
        self.assertEqual(force_str(response.content), "OK")

        # Martin can change Adrian's book
        self.assertTrue(self.client.login(username="martin", password="secr3t"))
        response = self.client.get(reverse("cbv.change_book", args=(1,)))
        self.assertEqual(response.status_code, 200)
        self.assertEqual(force_str(response.content), "OK")

        # Adrian can delete his book
        self.assertTrue(self.client.login(username="adrian", password="secr3t"))
        response = self.client.get(reverse("cbv.delete_book", args=(1,)))
        self.assertEqual(response.status_code, 200)
        self.assertEqual(force_str(response.content), "OK")

        # Martin can *not* delete Adrian's book
        # Up to Django v2.1, the response was a redirect to login
        self.assertTrue(self.client.login(username="martin", password="secr3t"))
        response = self.client.get(reverse("cbv.delete_book", args=(1,)))
        self.assertIn(response.status_code, [302, 403])

        # Martin can *not* delete Adrian's book and an PermissionDenied is raised
        self.assertTrue(self.client.login(username="martin", password="secr3t"))
        response = self.client.get(reverse("cbv.view_that_raises", args=(1,)))
        self.assertEqual(response.status_code, 403)

        # Test views that require a list of permissions

        # Adrian has both permissions
        self.assertTrue(self.client.login(username="adrian", password="secr3t"))
        response = self.client.get(reverse("cbv.view_with_permission_list", args=(1,)))
        self.assertEqual(response.status_code, 200)
        self.assertEqual(force_str(response.content), "OK")

        # Martin does not have delete permission
        # Up to Django v2.1, the response was a redirect to login
        self.assertTrue(self.client.login(username="martin", password="secr3t"))
        response = self.client.get(reverse("cbv.view_with_permission_list", args=(1,)))
        self.assertIn(response.status_code, [302, 403])


@unittest.skipIf(sys.version_info.major < 3, "Python 3 only")
class AutoPermissionRequiredMixinTests(TestCase):
    def setUp(self):
        from testapp.models import TestModel

        self.model = TestModel
        self.req = RequestFactory().get("/")
        self.req.user = AnonymousUser()

    def test_predefined_view_type(self):
        class TestView(AutoPermissionRequiredMixin, CreateView):
            model = self.model
            fields = ()

        self.assertEqual(TestView.as_view()(self.req).status_code, 200)

    def test_custom_view_type(self):
        class CustomViewMixin:
            pass

        class TestView(AutoPermissionRequiredMixin, CustomViewMixin, CreateView):
            model = self.model
            fields = ()
            permission_type_map = [
                (CustomViewMixin, "unknown_perm")
            ] + AutoPermissionRequiredMixin.permission_type_map
            raise_exception = True

        with self.assertRaises(PermissionDenied):
            TestView.as_view()(self.req)

    def test_unknown_view_type(self):
        class TestView(AutoPermissionRequiredMixin, View):
            pass

        with self.assertRaises(ImproperlyConfigured):
            TestView.as_view()(self.req)

    def test_overwrite_perm_type(self):
        class TestView(AutoPermissionRequiredMixin, CreateView):
            model = self.model
            fields = ()
            permission_type = "unknown"
            raise_exception = True

        with self.assertRaises(PermissionDenied):
            TestView.as_view()(self.req)

    def test_disable_perm_checking(self):
        class TestView(AutoPermissionRequiredMixin, CreateView):
            model = self.model
            fields = ()
            permission_type = None

        self.assertEqual(TestView.as_view()(self.req).status_code, 200)

    def test_permission_required_passthrough(self):
        class TestView(AutoPermissionRequiredMixin, CreateView):
            model = self.model
            fields = ()
            permission_required = "testapp.unknown_perm"
            raise_exception = True

        with self.assertRaises(PermissionDenied):
            TestView.as_view()(self.req)



from __future__ import absolute_import

import rules


# Predicates

@rules.predicate
def is_book_author(user, book):
    if not book:
        return False
    return book.author == user


@rules.predicate
def is_boss(user):
    return user.is_superuser


is_editor = rules.is_group_member('editors')

# Rules

rules.add_rule('change_book', is_book_author | is_editor)
rules.add_rule('delete_book', is_book_author)
rules.add_rule('create_book', is_boss)

# Permissions

rules.add_perm('testapp.change_book', is_book_author | is_editor)
rules.add_perm('testapp.delete_book', is_book_author)



from django.core.exceptions import ImproperlyConfigured
from django.db.models import Model
from django.db.models.base import ModelBase

from ..permissions import add_perm


class RulesModelBaseMixin:
    """
    Mixin for the metaclass of Django's Model that allows declaring object-level
    permissions in the model's Meta options.

    If set, the Meta attribute "rules_permissions" has to be a dictionary with
    permission types (like "add" or "change") as keys and predicates (like
    rules.is_staff) as values. Permissions are then registered with the rules
    framework automatically upon Model creation.

    This mixin can be used for creating custom metaclasses.
    """

    def __new__(cls, name, bases, attrs, **kwargs):
        model_meta = attrs.get("Meta")
        if hasattr(model_meta, "rules_permissions"):
            perms = model_meta.rules_permissions
            del model_meta.rules_permissions
            if not isinstance(perms, dict):
                raise ImproperlyConfigured(
                    "The rules_permissions Meta option of %s must be a dict, not %s."
                    % (name, type(perms))
                )
            perms = perms.copy()
        else:
            perms = {}

        new_class = super().__new__(cls, name, bases, attrs, **kwargs)
        new_class._meta.rules_permissions = perms
        new_class.preprocess_rules_permissions(perms)
        for perm_type, predicate in perms.items():
            add_perm(new_class.get_perm(perm_type), predicate)
        return new_class


class RulesModelBase(RulesModelBaseMixin, ModelBase):
    """
    A subclass of Django's ModelBase with the RulesModelBaseMixin mixed in.
    """


class RulesModelMixin:
    """
    A mixin for Django's Model that adds hooks for stepping into the process of
    permission registration, which are called by the metaclass implementation in
    RulesModelBaseMixin.

    Use this mixin in a custom subclass of Model in order to change its behavior.
    """

    @classmethod
    def get_perm(cls, perm_type):
        """Converts permission type ("add") to permission name ("app.add_modelname")

        :param perm_type: "add", "change", etc., or custom value
        :type  perm_type: str
        :returns str:
        """
        return "%s.%s_%s" % (cls._meta.app_label, perm_type, cls._meta.model_name)

    @classmethod
    def preprocess_rules_permissions(cls, perms):
        """May alter a permissions dict before it's processed further.

        Use this, for instance, to alter the supplied permissions or insert default
        values into the given dict.

        :param perms:
            Shallow-copied value of the rules_permissions model Meta option
        :type  perms: dict
        """


class RulesModel(RulesModelMixin, Model, metaclass=RulesModelBase):
    """
    An abstract model with RulesModelMixin mixed in, using RulesModelBase as metaclass.

    Use this as base for your models directly if you don't need to customize the
    behavior of RulesModelMixin and thus don't want to create a custom base class.
    """

    class Meta:
        abstract = True



from unittest import TestCase

from rules.predicates import always_true, always_false
from rules.permissions import (permissions, add_perm, set_perm, remove_perm,
                               perm_exists, has_perm, ObjectPermissionBackend)


class PermissionsTests(TestCase):
    @staticmethod
    def reset_ruleset(ruleset):
        for k in list(ruleset.keys()):
            ruleset.pop(k)

    def setUp(self):
        self.reset_ruleset(permissions)

    def tearDown(self):
        self.reset_ruleset(permissions)

    def test_permissions_ruleset(self):
        add_perm('can_edit_book', always_true)
        assert 'can_edit_book' in permissions
        assert perm_exists('can_edit_book')
        assert has_perm('can_edit_book')
        with self.assertRaises(KeyError):
            add_perm('can_edit_book', always_false)
        set_perm('can_edit_book', always_false)
        assert not has_perm('can_edit_book')
        remove_perm('can_edit_book')
        assert not perm_exists('can_edit_book')

    def test_backend(self):
        backend = ObjectPermissionBackend()
        assert backend.authenticate('someuser', 'password') is None

        add_perm('can_edit_book', always_true)
        assert 'can_edit_book' in permissions
        assert backend.has_perm(None, 'can_edit_book')
        assert backend.has_module_perms(None, 'can_edit_book')
        with self.assertRaises(KeyError):
            add_perm('can_edit_book', always_true)
        set_perm('can_edit_book', always_false)
        assert not backend.has_perm(None, 'can_edit_book')
        remove_perm('can_edit_book')
        assert not perm_exists('can_edit_book')



from django.contrib.auth.models import User, Group
from testapp.models import Book

import testapp.rules  # to register rules


ISBN = '978-1-4302-1936-1'


class TestData:
    @classmethod
    def setUpTestData(cls):
        adrian = User.objects.create_user(
            'adrian',
            password='secr3t',
            is_superuser=True,
            is_staff=True)

        martin = User.objects.create_user(
            'martin',
            password='secr3t',
            is_staff=True)

        editors = Group.objects.create(name='editors')
        martin.groups.add(editors)

        Book.objects.create(
            isbn=ISBN,
            title='The Definitive Guide to Django',
            author=adrian)



from unittest import TestCase

from rules.predicates import always_true, always_false
from rules.rulesets import (RuleSet, default_rules, add_rule, set_rule,
                            remove_rule, rule_exists, test_rule)


class RulesetTests(TestCase):
    @staticmethod
    def reset_ruleset(ruleset):
        for k in list(ruleset.keys()):
            ruleset.pop(k)

    def setUp(self):
        self.reset_ruleset(default_rules)

    def tearDown(self):
        self.reset_ruleset(default_rules)

    def test_shared_ruleset(self):
        add_rule('somerule', always_true)
        assert 'somerule' in default_rules
        assert rule_exists('somerule')
        assert test_rule('somerule')
        assert test_rule('somerule')
        with self.assertRaises(KeyError):
            add_rule('somerule', always_false)
        set_rule('somerule', always_false)
        assert not test_rule('somerule')
        remove_rule('somerule')
        assert not rule_exists('somerule')

    def test_ruleset(self):
        ruleset = RuleSet()
        ruleset.add_rule('somerule', always_true)
        assert 'somerule' in ruleset
        assert ruleset.rule_exists('somerule')
        assert ruleset.test_rule('somerule')
        with self.assertRaises(KeyError):
            ruleset.add_rule('somerule', always_true)
        ruleset.set_rule('somerule', always_false)
        assert not test_rule('somerule')
        ruleset.remove_rule('somerule')
        assert not ruleset.rule_exists('somerule')



# Generated by Django 2.2.2 on 2019-08-05 08:04

import sys

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='Book',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('isbn', models.CharField(max_length=50, unique=True)),
                ('title', models.CharField(max_length=100)),
                ('author', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to=settings.AUTH_USER_MODEL)),
            ],
        ),
    ]

    # TestModel doesn't work under Python 2
    if sys.version_info.major >= 3:
        import rules.contrib.models
        operations += [
            migrations.CreateModel(
                name='TestModel',
                fields=[
                    ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ],
                bases=(rules.contrib.models.RulesModelMixin, models.Model),
            ),
        ]



#!/usr/bin/env python
from os.path import dirname, join

try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup

from rules import VERSION


def get_version(version):
    """
    Returns a PEP 386-compliant version number from ``version``.
    """
    assert len(version) == 5
    assert version[3] in ('alpha', 'beta', 'rc', 'final')

    parts = 2 if version[2] == 0 else 3
    main = '.'.join(str(x) for x in version[:parts])

    sub = ''
    if version[3] != 'final':
        mapping = {'alpha': 'a', 'beta': 'b', 'rc': 'c'}
        sub = mapping[version[3]] + str(version[4])

    return main + sub


with open(join(dirname(__file__), 'README.rst')) as f:
    long_description = f.read()


setup(
    name='rules',
    description='Awesome Django authorization, without the database',
    version=get_version(VERSION),
    long_description=long_description,

    url='http://github.com/dfunckt/django-rules',
    author='Akis Kesoglou',
    author_email='akiskesoglou@gmail.com',
    maintainer='Akis Kesoglou',
    maintainer_email='akiskesoglou@gmail.com',
    license='MIT',

    zip_safe=False,
    packages=[
        'rules',
        'rules.templatetags',
        'rules.contrib',
        'rules.compat',
    ],

    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Web Environment',
        'Framework :: Django',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
    ],
)



from django.test import TestCase
from django.urls import reverse

from . import TestData


class ModelAdminTests(TestData, TestCase):
    def test_change_book(self):
        # adrian can change his book as its author
        self.assertTrue(self.client.login(username='adrian', password='secr3t'))
        response = self.client.get(reverse('admin:testapp_book_change', args=(1,)))
        self.assertEqual(response.status_code, 200)

        # martin can change adrian's book as an editor
        self.assertTrue(self.client.login(username='martin', password='secr3t'))
        response = self.client.get(reverse('admin:testapp_book_change', args=(1,)))
        self.assertEqual(response.status_code, 200)

    def test_delete_book(self):
        # martin can *not* delete adrian's book
        self.assertTrue(self.client.login(username='martin', password='secr3t'))
        response = self.client.get(reverse('admin:testapp_book_delete', args=(1,)))
        self.assertEqual(response.status_code, 403)

        # adrian can delete his book as its author
        self.assertTrue(self.client.login(username='adrian', password='secr3t'))
        response = self.client.get(reverse('admin:testapp_book_delete', args=(1,)))
        self.assertEqual(response.status_code, 200)



from django.contrib.auth.models import User
from django.test import TestCase

from rules.predicates import (is_authenticated, is_superuser, is_staff,
                              is_active, is_group_member)

from . import TestData


class SwappedUser(object):
    pass


class PredicateTests(TestData, TestCase):
    def test_is_authenticated(self):
        assert is_authenticated(User.objects.get(username='adrian'))
        assert not is_authenticated(SwappedUser())

    def test_is_superuser(self):
        assert is_superuser(User.objects.get(username='adrian'))
        assert not is_superuser(SwappedUser())

    def test_is_staff(self):
        assert is_staff(User.objects.get(username='adrian'))
        assert not is_staff(SwappedUser())

    def test_is_active(self):
        assert is_active(User.objects.get(username='adrian'))
        assert not is_active(SwappedUser())

    def test_is_group_member(self):
        p1 = is_group_member('somegroup')
        assert p1.name == 'is_group_member:somegroup'
        assert p1.num_args == 1

        p2 = is_group_member('g1', 'g2', 'g3', 'g4')
        assert p2.name == 'is_group_member:g1,g2,g3,...'

        p = is_group_member('editors')
        assert p(User.objects.get(username='martin'))
        assert not p(SwappedUser())

        p = is_group_member('editors', 'staff')
        assert not p(User.objects.get(username='martin'))
        assert not p(SwappedUser())



from django.contrib.auth.models import User
from django.template import Template, Context
from django.test import TestCase

from testapp.models import Book

from . import ISBN, TestData


class TemplateTagTests(TestData, TestCase):
    tpl_format = """{{% spaceless %}}
        {{% load rules %}}
        {{% {tag} "{name}" user book as can_update %}}
        {{% if can_update %}}
        OK
        {{% else %}}
        NOT OK
        {{% endif %}}
        {{% endspaceless %}}"""

    def test_rule_tag(self):

        # change_book rule

        tpl = Template(self.tpl_format.format(tag='test_rule', name='change_book'))

        # adrian can change his book as its author
        html = tpl.render(Context({
            'user': User.objects.get(username='adrian'),
            'book': Book.objects.get(isbn=ISBN),
        }))
        self.assertEqual(html, 'OK')

        # martin can change adrian's book as an editor
        html = tpl.render(Context({
            'user': User.objects.get(username='martin'),
            'book': Book.objects.get(isbn=ISBN),
        }))
        self.assertEqual(html, 'OK')

        # delete_book rule

        tpl = Template(self.tpl_format.format(tag='test_rule', name='delete_book'))

        # adrian can delete his book as its author
        html = tpl.render(Context({
            'user': User.objects.get(username='adrian'),
            'book': Book.objects.get(isbn=ISBN),
        }))
        self.assertEqual(html, 'OK')

        # martin can *not* delete adrian's book
        html = tpl.render(Context({
            'user': User.objects.get(username='martin'),
            'book': Book.objects.get(isbn=ISBN),
        }))
        self.assertEqual(html, 'NOT OK')

    def test_perm_tag(self):

        # change_book permission

        tpl = Template(self.tpl_format.format(tag='has_perm', name='testapp.change_book'))

        # adrian can change his book as its author
        html = tpl.render(Context({
            'user': User.objects.get(username='adrian'),
            'book': Book.objects.get(isbn=ISBN),
        }))
        self.assertEqual(html, 'OK')

        # martin can change adrian's book as an editor
        html = tpl.render(Context({
            'user': User.objects.get(username='martin'),
            'book': Book.objects.get(isbn=ISBN),
        }))
        self.assertEqual(html, 'OK')

        # delete_book permission

        tpl = Template(self.tpl_format.format(tag='has_perm', name='testapp.delete_book'))

        # adrian can delete his book as its author
        html = tpl.render(Context({
            'user': User.objects.get(username='adrian'),
            'book': Book.objects.get(isbn=ISBN),
        }))
        self.assertEqual(html, 'OK')

        # martin can *not* delete adrian's book
        html = tpl.render(Context({
            'user': User.objects.get(username='martin'),
            'book': Book.objects.get(isbn=ISBN),
        }))
        self.assertEqual(html, 'NOT OK')



from __future__ import absolute_import

from django.contrib import admin
from rules.contrib.admin import ObjectPermissionsModelAdmin
from .models import Book


class BookAdmin(ObjectPermissionsModelAdmin):
    pass


admin.site.register(Book, BookAdmin)
